import base58  # type: ignore
import re

from bech32 import bech32_encode, bech32_decode, CHARSET  # type: ignore
from binascii import unhexlify
from bitstring import ConstBitStream  # type: ignore
from ecdsa import SECP256k1, VerifyingKey  # type: ignore
from ecdsa.util import sigdecode_string  # type: ignore
from hashlib import sha256

from .types import LightningInvoice, MilliSatoshi, Route, Signature
from .utils import amount_to_msat, trim_to_bytes, bitarray_to_u5, u5_to_bitarray


base58_prefix_map = {"bc": (0, 5), "tb": (111, 196)}


def _pull_tagged(stream):
    tag = stream.read(5).uint
    length = stream.read(5).uint * 32 + stream.read(5).uint

    return (CHARSET[tag], stream.read(length * 5), stream)


def _parse_fallback(fallback, currency):
    if currency in ["bc", "tb"]:
        wver = fallback[0:5].uint
        if wver == 17:
            addr = base58.b58encode_check(bytes([base58_prefix_map[currency][0]]) + fallback[5:].tobytes())
        elif wver == 18:
            addr = base58.b58encode_check(bytes([base58_prefix_map[currency][1]]) + fallback[5:].tobytes())
        elif wver <= 16:
            addr = bech32_encode(currency, bitarray_to_u5(fallback))
        else:
            return None
    else:
        addr = fallback.tobytes()
    return addr


def _readable_scid(short_channel_id: int) -> str:
    return "{blockheight}x{transactionindex}x{outputindex}".format(
        blockheight=((short_channel_id >> 40) & 0xFFFFFF),
        transactionindex=((short_channel_id >> 16) & 0xFFFFFF),
        outputindex=(short_channel_id & 0xFFFF),
    )


def encode():
    raise NotImplementedError


def decode(bech32_pr: str) -> LightningInvoice:
    hrp, bech32_data = bech32_decode(bech32_pr)
    route_hints = []
    tags = {}

    if not hrp or not bech32_data or not hrp.startswith("ln"):
        raise ValueError("Bech32 is not valid.")

    matches = re.match(r"ln(bcrt|bc|tb)(\w+)?", hrp)
    assert matches, "Human readable part is not valid."

    currency, amount_str = matches.groups()
    data_part = u5_to_bitarray(bech32_data)

    # Final signature is 65 bytes, split it off.
    # signature =>
    # "a valid 512-bit secp256k1 signature of the SHA2 256-bit hash of the human-readable part
    # represented as UTF-8 bytes, concatenated with the data part (excluding the signature) with 0 bits appended
    # to pad the data to the next byte boundary, with a trailing byte containing the recovery ID (0, 1, 2, or 3)"

    if len(data_part) < 65 * 8:
        raise ValueError("Too short to contain signature")

    signature_data = data_part[-65 * 8 :].tobytes()
    data = ConstBitStream(data_part[: -65 * 8])
    signature = Signature(data=signature_data, signing_data=(hrp.encode("utf-8") + data.tobytes()))
    timestamp = data.read(35).uint

    # Look for tags in data

    while data.pos != data.len:
        tag, tagdata, data = _pull_tagged(data)
        data_length = len(tagdata) // 5

        if tag == "p" and data_length == 52:
            # p (1): data_length 52. 256-bit SHA256 payment_hash. Preimage of this provides proof of payment.
            tags["p"] = trim_to_bytes(tagdata).hex()

        elif tag == "x":
            # x (6): data_length variable. expiry time in seconds (big-endian). Default is 3600.
            tags["x"] = tagdata.uint

        elif tag == "d":
            # d (13): data_length variable. Short description of purpose of payment (UTF-8).
            tags["d"] = trim_to_bytes(tagdata).decode("utf-8")

        elif tag == "h" and data_length == 52:
            # h (23): data_length 52. 256-bit description of purpose of payment (SHA256).
            tags["h"] = trim_to_bytes(tagdata).hex()

        elif tag == "s" and data_length == 52:
            # s (16): data_length 52. This 256-bit secret prevents forwarding nodes from probing the payment recipient.
            tags["s"] = trim_to_bytes(tagdata).hex()

        elif tag == "c":
            # c (24): data_length variable. min_final_cltv_expiry to use for the last HTLC in the route. Default is 9.
            tags["c"] = tagdata.uint

        elif tag == "n" and data_length == 53:
            # n (19): data_length 53. 33-byte public key of the payee node.
            tags["n"] = trim_to_bytes(tagdata).hex()

        elif tag == "f":
            # f (9): data_length variable, depending on version. Fallback on-chain address.
            tags["f"] = _parse_fallback(tagdata, currency)

        elif tag == "r":
            # r (3): `data_length` variable.
            # One or more entries containing extra routing information for a private route;
            # there may be more than one `r` field, too.
            #  * `pubkey` (264 bits)
            #  * `short_channel_id` (64 bits)
            #  * `feebase` (32 bits, big-endian)
            #  * `feerate` (32 bits, big-endian)
            #  * `cltv_expiry_delta` (16 bits, big-endian)
            s = ConstBitStream(tagdata)

            while s.pos + 264 + 64 + 32 + 32 + 16 < s.len:
                route_hints.append(
                    Route(
                        public_key=s.read(264).tobytes().hex(),
                        short_channel_id=_readable_scid(s.read(64).intbe),
                        base_fee=MilliSatoshi(s.read(32).intbe),
                        ppm_fee=s.read(32).intbe,
                        cltv_expiry_delta=s.read(16).intbe,
                    )
                )

    # A reader MUST use the `n` field to validate the signature instead of
    # performing signature recovery if a valid `n` field is provided.

    message = bytearray([ord(c) for c in hrp]) + data.tobytes()
    sig = signature_data[0:64]

    if "n" in tags:
        payee_public_key = tags["n"]
        vk = VerifyingKey.from_string(unhexlify(payee_public_key), curve=SECP256k1)
        if not vk.verify(sig, message, sha256, sigdecode=sigdecode_string):
            raise ValueError("Could not verify public key")
    else:
        vk = VerifyingKey.from_public_key_recovery(sig, message, SECP256k1, sha256)
        signaling_byte = signature_data[64]
        key = vk[int(signaling_byte)]
        payee_public_key = key.to_string("compressed").hex()

    return LightningInvoice(
        amount=amount_to_msat(amount_str) if amount_str else None,
        currency=currency,
        timestamp=timestamp,
        payee_public_key=payee_public_key,
        route_hints=route_hints,
        signature=signature,
        tags=tags,
    )


import time

from datetime import datetime
from decimal import Decimal
from hashlib import sha256
from typing import Any, Dict, List, NamedTuple, Optional


class LightningInvoice(NamedTuple):
    """Lightning invoice."""

    currency: str
    payee_public_key: str
    signature: "Signature"
    tags: Dict[str, Any]
    timestamp: int
    amount: Optional["MilliSatoshi"] = None
    route_hints: List["Route"] = []

    @property
    def description(self) -> Optional[str]:
        return self.tags["d"] if "d" in self.tags else None

    @property
    def description_hash(self) -> Optional[str]:
        return self.tags["h"] if "h" in self.tags else None

    @property
    def dt(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp)

    @property
    def expiry_time(self) -> int:
        return self.tags["x"] if "x" in self.tags else 3600

    @property
    def fallback_on_chain_address(self) -> Optional[str]:
        return self.tags["f"] if "f" in self.tags else None

    @property
    def min_final_cltv_expiry(self) -> Optional[int]:
        return self.tags["c"] if "c" in self.tags else 9

    @property
    def payment_hash(self) -> str:
        return self.tags["p"]

    @property
    def payment_secret(self) -> Optional[str]:
        return self.tags["s"] if "s" in self.tags else None

    def has_expired(self) -> bool:
        return time.time() > self.timestamp + self.expiry_time

    def is_mainnet(self) -> bool:
        return self.currency == "bc"

    def is_testnet(self) -> bool:
        return self.currency == "tb"

    def is_regtest(self) -> bool:
        return self.currency == "bcrt"


class MilliSatoshi(int):
    """A thousandth of a satoshi."""

    @classmethod
    def from_btc(cls, btc: Decimal) -> "MilliSatoshi":
        return cls(btc * 100_000_000_000)

    @property
    def btc(self) -> Decimal:
        return Decimal(self) / 100_000_000_000

    @property
    def sat(self) -> int:
        return self // 1000


class Route(NamedTuple):
    public_key: str
    short_channel_id: str
    base_fee: MilliSatoshi
    ppm_fee: int
    cltv_expiry_delta: int


class Signature(NamedTuple):
    """An invoice signature."""

    data: bytes
    signing_data: bytes

    @property
    def r(self) -> str:
        return self.data[:32].hex()

    @property
    def s(self) -> str:
        return self.data[32:64].hex()

    @property
    def recovery_flag(self) -> int:
        return int(self.data[64])

    @property
    def preimage(self) -> bytes:
        return sha256(self.signing_data).digest()

    def hex(self) -> str:
        return self.data[:64].hex()


import asyncio
from fastapi import HTTPException
from starlette.requests import Request
from pymacaroons import Macaroon, Verifier
import datetime
import hashlib
import uuid
from vendor.bolt11.core import decode

class L402Validator:
    """
    Middleware for validating L402 (Lightning Service Authentication Tokens) in a FastAPI application.
    
    Attributes:
        root_key: The root key for generating and verifying macaroons.
        price_sats: The price (in satoshis) for accessing the API endpoint.
        expiry_sec: The expiration time (in seconds) for the macaroon.
        generate_invoice: A function for generating Lightning Network invoices.
        verified_macaroons: A dictionary for caching verified macaroons to ensure they are used only once.
        cleanup_interval: The interval (in seconds) at which expired macaroons are cleaned up from the cache.
    """
    def __init__(self, root_key: str, price_sats: int, expiry_sec: int, generate_invoice_func):
        """
        Initializes the L402Validator with the root key, price, expiry time, and invoice generation function.
        Starts the cleanup task for expired macaroons.
        """
        
        self.root_key = root_key
        self.price_sats = price_sats
        self.expiry_sec = expiry_sec
        self.generate_invoice = generate_invoice_func
        self.verified_macaroons = {}  # Cache to store verified macaroons
        self.cleanup_interval = 60 * 10  # Cleanup interval in seconds (10 minutes)
        asyncio.create_task(self.cleanup_expired_macaroons())  # Start the cleanup task

    async def cleanup_expired_macaroons(self):
        """
        Periodically cleanup expired macaroons from the cache.
        """
        while True:
            await asyncio.sleep(self.cleanup_interval)  # Wait for the cleanup interval
            # Create a list of expired macaroons
            expired_macaroons = [id for id, expiry in self.verified_macaroons.items() if datetime.datetime.now().isoformat() > expiry]
            # Remove expired macaroons from the cache
            for id in expired_macaroons:
                del self.verified_macaroons[id]

    
    async def __call__(self, request: Request):
        l402_key = request.headers.get('Authorization')
        if l402_key is None or not l402_key.startswith('LSAT '):

            try:
                label = str(uuid.uuid4()) 
                invoice = await self.generate_invoice(self.price_sats, f'LSAT_{label}', f'payment for endpoint')  # function to generate invoice
            except Exception as e:
                raise HTTPException(status_code=500, detail=f"Couldn't generate invoice. Reason: {str(e)}")

            encoded_invoice = invoice['bolt11']    
            decoded_invoice = decode(encoded_invoice)
            payment_hash = decoded_invoice.tags['p']
            print(payment_hash)
            macaroon = Macaroon(identifier=payment_hash, key=self.root_key)

            expiry_time = datetime.datetime.now() + datetime.timedelta(seconds=self.expiry_sec)
            macaroon.add_first_party_caveat(f"expires = {expiry_time.isoformat()}")
            macaroon.add_first_party_caveat(f"payment_hash = {payment_hash}")
            encoded_macaroon = macaroon.serialize()
            
            raise HTTPException(
                status_code=402,
                detail="Payment Required",
                headers={"content-type": "application/json",
                         "www-authenticate": f'LSAT macaroon="{encoded_macaroon}",invoice="{encoded_invoice}"'}
            )

        try:
            macaroon, preimage = l402_key.lstrip('LSAT ').split(':')
        except ValueError:
            raise HTTPException(status_code=400, detail="Invalid L402 key format")

        try:
            macaroon = Macaroon.deserialize(macaroon)
        except Exception as e:
            raise HTTPException(status_code=400, detail=f"Couldn't deserialize macaroon. Reason: {str(e)}")

        verifier = Verifier()
        
        verifier.satisfy_general(lambda caveat: caveat.split(' = ')[0] == 'expires' and datetime.datetime.now().isoformat() <= caveat.split(' = ')[1])
        print(hashlib.sha256(bytes.fromhex(preimage)).hexdigest())
        verifier.satisfy_general(lambda caveat: caveat.split(' = ')[0] == 'payment_hash' and hashlib.sha256(bytes.fromhex(preimage)).hexdigest() == caveat.split(' = ')[1])
        try:
            if not verifier.verify(macaroon, self.root_key):
                raise HTTPException(status_code=403, detail="Forbidden: Invalid macaroon or preimage")
        except Exception as e:
            raise HTTPException(status_code=403, detail=f"Forbidden: {str(e)}")

        if macaroon.identifier in self.verified_macaroons:
            raise HTTPException(status_code=403, detail=f"Forbidden: Macaroon is only valid once")

        # If verification is successful, add macaroon to cache with its expiry time
        for caveat in macaroon.caveats:
            if caveat.caveat_id.split(' = ')[0] == 'expires':
                self.verified_macaroons[macaroon.identifier] = caveat.caveat_id.split(' = ')[1]

        return request  # Add this line
